from pathlib import Path
import unittest
from shutil import which
from sotools import is_elf, library_links
from sotools.ldd import ldd
from sotools.libraryset import LibrarySet, Library
from sotools.linker import resolve


class ToolsTest(unittest.TestCase):

    @unittest.skipIf(not which('ls'), "No binary to test with")
    def test_ldd(self):
        ls_bin = which('ls')
        libraries = ldd(ls_bin)

        self.assertIn('libc.so.6', libraries.sonames)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_library_links(self):
        target = Library.from_path(resolve('libm.so.6'))

        links = library_links(target)

        for path in links:
            self.assertEqual(
                Path(path).resolve(),
                Path(target.binary_path).resolve())

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_is_elf(self):
        self.assertFalse(is_elf('/proc/meminfo'))
        self.assertFalse(is_elf('/'))
        self.assertTrue(is_elf(resolve('libm.so.6')))



import unittest
from pathlib import Path
from sotools.dl_cache.structure import (BinaryStruct,
                                        deserialize_null_terminated_string)
from sotools.dl_cache.extensions import (CacheExtension, CacheExtensionSection,
                                         CacheExtensionTag,
                                         cache_extension_sections)
from sotools.dl_cache.dl_cache import (_cache_type, _CacheType, _CacheHeader,
                                       _CacheHeaderNew, _CacheHeaderOld,
                                       _FileEntryNew, _FileEntryOld)
from sotools.dl_cache.extensions.hwcaps import (dl_cache_hwcap_extension,
                                                HWCAPSection)
from sotools.dl_cache.flags import Flags
from sotools.dl_cache import (cache_libraries, _cache_libraries, get_generator,
                              search_cache, _parse_cache, DynamicLinkerCache)

EMBEDDED_CACHE = f'{Path(__file__).parent}/assets/embedded.so.cache'
MODERN_CACHE = f'{Path(__file__).parent}/assets/modern.so.cache'
HWCAPS_CACHE = f'{Path(__file__).parent}/assets/with_hwcaps.so.cache'


class DLCacheTest(unittest.TestCase):

    def test_structure_bad_format(self):

        class MyStruct(BinaryStruct):
            pass

        with self.assertRaises(NotImplementedError):
            MyStruct.deserialize("Dummy string".encode())

        with self.assertRaises(NotImplementedError):
            BinaryStruct.sizeof(MyStruct)

    def test_format_detect(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            type_, offset = _cache_type(cache_file.read())

        self.assertEqual(type_, _CacheType.NEW_FORMAT)
        self.assertEqual(offset, 0)

        with open(EMBEDDED_CACHE, 'rb') as cache_file:
            type_, offset = _cache_type(cache_file.read())

        self.assertEqual(type_, _CacheType.NEW_FORMAT)
        self.assertNotEqual(offset, 0)

        type_, offset = _cache_type("This is not a cache".encode())

        self.assertEqual(type_, _CacheType.UNKNOWN)

    def test_class_routing(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        header = _CacheHeader.deserialize(cache_data)
        self.assertTrue(isinstance(header, _CacheHeaderNew))

        with open(EMBEDDED_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        _, offset = _cache_type(cache_data)

        header = _CacheHeader.deserialize(cache_data[:offset])
        self.assertTrue(isinstance(header, _CacheHeaderOld))

        header = _CacheHeader.deserialize(cache_data)
        self.assertTrue(isinstance(header, _CacheHeaderNew))

    def test_generated_repr(self):
        for cls in (_CacheHeaderNew, _CacheHeaderOld, _FileEntryNew,
                    _FileEntryOld):
            self.assertIn(cls.__name__, str(cls()))

            for (attribute, value) in cls.__structure__:
                if attribute is not None:
                    self.assertIn(attribute, str(cls()))

    def test_deserialize_bad_header(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        for i in range(24):
            with self.assertRaises(Exception):
                header = _CacheHeader.deserialize(cache_data[:i])

    def test_list_libraries(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        self.assertTrue(_cache_libraries(cache_data))

        with open(EMBEDDED_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        self.assertTrue(_cache_libraries(cache_data))

    def test_list_libraries_bad_cache(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        with self.assertRaises(Exception):
            libs = _cache_libraries(cache_data[:22])

        with self.assertRaises(Exception):
            libs = _cache_libraries(cache_data[:100])

    def test_wrapper(self):
        self.assertFalse(
            cache_libraries(cache_file=MODERN_CACHE,
                            arch_flags=Flags.FLAG_SPARC_LIB64))
        self.assertTrue(cache_libraries(cache_file=MODERN_CACHE))
        self.assertTrue(cache_libraries(cache_file=EMBEDDED_CACHE))

    def test_structure_sizeof(self):

        class TestByte(BinaryStruct):
            __structure__ = [('member', 'uint8_t')]

        class TestBytes(BinaryStruct):
            __structure__ = [('member', 'uint8_t'), ('other', 'uint8_t')]

        class TestInt(BinaryStruct):
            __structure__ = [('member', 'uint32_t')]

        class TestLong(BinaryStruct):
            __structure__ = [('member', 'uint64_t')]

        self.assertEqual(BinaryStruct.sizeof(TestByte), 1)
        self.assertEqual(BinaryStruct.sizeof(TestBytes), 2)
        self.assertEqual(BinaryStruct.sizeof(TestInt), 4)
        self.assertEqual(BinaryStruct.sizeof(TestLong), 8)

        class TestPadding(BinaryStruct):
            __structure__ = [('member', 'uint64_t'), (None, 24),
                             ('other', 'uint32_t')]

        self.assertEqual(BinaryStruct.sizeof(TestPadding), 36)

        class TestPaddingBadFormat(BinaryStruct):
            __structure__ = [('member', 'uint64_t'), (None, None),
                             ('other', 'uint32_t')]

        self.assertEqual(BinaryStruct.sizeof(TestPaddingBadFormat), 12)

    def test_structure_deserialize(self):

        class TestStruct(BinaryStruct):
            __structure__ = [('a', 'uint32_t'), ('b', 'uint8_t')]

        parsed = TestStruct.deserialize(
            int(10).to_bytes(4, 'big') + int(254).to_bytes(1, 'big'))
        self.assertIsInstance(parsed, TestStruct)

        with self.assertRaises(NotImplementedError):
            parsed = TestStruct.deserialize("test")

    def test_deserialize_header(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        header = _CacheHeader.deserialize(cache_data)

        self.assertTrue(header.nlibs)

    def test_deserialize_entry(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        header = _CacheHeader.deserialize(cache_data)
        header_size = BinaryStruct.sizeof(header.__class__)
        entry_type = header.__class__.entry_type

        entry = entry_type.deserialize(cache_data[header_size:])
        self.assertNotEqual(entry.key, 0)
        self.assertNotEqual(entry.value, 0)
        self.assertNotEqual(entry.flags, 0)

        self.assertLess(entry.key, header.extension_offset)
        self.assertLess(entry.value, header.extension_offset)

    def test_deserialize_extension_header(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        header = _CacheHeader.deserialize(cache_data)
        extension_header = CacheExtension.deserialize(
            cache_data[header.extension_offset:])

        self.assertNotEqual(extension_header.count, 0)

    def test_deserialize_extension_sections(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        header = _CacheHeader.deserialize(cache_data)

        for section in cache_extension_sections(
                cache_data[header.offset + header.extension_offset:]):
            self.assertIn(
                section.tag, {
                    CacheExtensionTag.TAG_GENERATOR,
                    CacheExtensionTag.TAG_GLIBC_HWCAPS
                })
            self.assertNotEqual(section.offset, 0)
            self.assertNotEqual(section.size, 0)

    def test_deserialize_nts(self):
        string = "I am a string"
        data = f"{string}\0".encode()

        self.assertEqual(deserialize_null_terminated_string(data), string)
        self.assertEqual(deserialize_null_terminated_string(string.encode()),
                         "")

    def test_get_hwcap_string(self):
        with open(HWCAPS_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        header = _CacheHeader.deserialize(cache_data)

        for section in cache_extension_sections(
                cache_data[header.offset + header.extension_offset:]):
            if section.tag == CacheExtensionTag.TAG_GLIBC_HWCAPS:
                hwcap_section = HWCAPSection(section)
                self.assertTrue(hwcap_section.string_value(cache_data))

        badlength = HWCAPSection(CacheExtensionSection())
        HWCAPSection.offset = 0
        HWCAPSection.size = 40

        self.assertEqual(
            badlength.string_value("Not an int, too short".encode()), "")

    def test_assert_hwcap_reference(self):
        entry = _FileEntryNew()

        entry.hwcap = 4611686018427387904
        self.assertTrue(dl_cache_hwcap_extension(entry))

        entry.hwcap = 0
        self.assertFalse(dl_cache_hwcap_extension(entry))

        entry.hwcap = 1
        self.assertFalse(dl_cache_hwcap_extension(entry))

        entry = _FileEntryOld()
        self.assertFalse(dl_cache_hwcap_extension(entry))

    def test_parse_hwcap(self):
        cache = _parse_cache(HWCAPS_CACHE)

        self.assertTrue(set(filter(lambda x: x.hwcaps, cache.entries)))

    def test_extension_generator(self):
        with open(MODERN_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        self.assertTrue(get_generator(cache_data))

        with open(EMBEDDED_CACHE, 'rb') as cache_file:
            cache_data = cache_file.read()

        self.assertIsNone(get_generator(cache_data))

    def test_search_cache(self):
        self.assertTrue(search_cache('libc.so.6', MODERN_CACHE))
        self.assertIsNone(search_cache('notalib.so', MODERN_CACHE))
        self.assertIsNone(
            search_cache('libc.so.6', MODERN_CACHE, Flags.FLAG_POWERPC_LIB64))

    def test_parse_cache(self):
        self.assertIsNone(_parse_cache('/not/a/file'))
        self.assertIsNone(_parse_cache(__file__))
        self.assertIsInstance(_parse_cache(EMBEDDED_CACHE), DynamicLinkerCache)
        self.assertIsInstance(_parse_cache(MODERN_CACHE), DynamicLinkerCache)
        self.assertIsInstance(_parse_cache(HWCAPS_CACHE), DynamicLinkerCache)



#!/bin/env python3

import sys
import logging
from argparse import ArgumentParser
from sotools.dl_cache import _cache_libraries, Flags, get_generator

DEFAULT_CACHE = "/etc/ld.so.cache"
DESCRIPTION = """List the contents of a given dynamic linker cache."""
EPILOG = """Please report any mismatch between the dynamic linker config tool and the output of this program to http://github.com/spoutn1k/python-sotools."""

PARSER = ArgumentParser(
    prog='ldconfig.py',
    description=DESCRIPTION,
    epilog=EPILOG,
)

PARSER.add_argument(
    "cache",
    help="Path to an dynamic linker cache to dump.",
    nargs='?',
    default=DEFAULT_CACHE,
)

PARSER.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="Toggle verbose output",
)


def main():
    args = PARSER.parse_args()

    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
        )

    try:
        with open(args.cache, 'rb') as cache_file:
            cache_data = cache_file.read()
    except Exception as err:
        print(err, file=sys.stderr)
        sys.exit(1)

    libs = _cache_libraries(cache_data)
    print(f"{len(libs)} libs found in cache `{args.cache}'")

    for library in libs:
        hwcap_entry_string = f" hwcap: \"{library.hwcaps}\"" if library.hwcaps else ""
        description = ",".join(
            filter(None,
                   [Flags.description(library.flags), hwcap_entry_string]))
        print(f"\t{library.key} ({description}) => {library.value}")

    generator = get_generator(cache_data)
    if generator:
        print(f"Cache generated by: {generator}")

    sys.exit(0)



import struct
import logging
from sotools.dl_cache.structure import (DATATYPES,
                                        deserialize_null_terminated_string)
from sotools.dl_cache.extensions import CacheExtensionSection

# This bit in the hwcap field of struct file_entry_new indicates that
# the lower 32 bits contain an index into the
# cache_extension_tag_glibc_hwcaps section.  Older glibc versions do
# not know about this HWCAP bit, so they will ignore these
# entries.
DL_CACHE_HWCAP_EXTENSION = (1 << 62)

# The number of the ISA level bits in the upper 32 bits of the hwcap
# field.
DL_CACHE_HWCAP_ISA_LEVEL_COUNT = 10

# The mask of the ISA level bits in the hwcap field.
DL_CACHE_HWCAP_ISA_LEVEL_MASK = ((1 << DL_CACHE_HWCAP_ISA_LEVEL_COUNT) - 1)


def dl_cache_hwcap_extension(entry):
    hwcap_field = getattr(entry, 'hwcap', None)

    if hwcap_field is None:
        return False

    return ((hwcap_field >> 32) & ~DL_CACHE_HWCAP_ISA_LEVEL_MASK) == (
        DL_CACHE_HWCAP_EXTENSION >> 32)


class HWCAPSection(CacheExtensionSection):

    def __init__(self, section=None):
        super().__init__()

        # Copy to self the fields defined by the class
        for attribute, type_ in CacheExtensionSection.__structure__:
            # Fetch the default for this attribute type
            _, _, default = DATATYPES.get(type_, int)

            # Access the field value then set it in self
            value = getattr(section, attribute, default())
            setattr(self, attribute, value)

    def string_value(self, data):
        hwcap_data = data[self.offset:self.offset + self.size]

        try:
            hwcap_pointer, = struct.unpack("I", hwcap_data)
        except struct.error as err:
            logging.error("Failed to retrieve hwcap string value: %s",
                          str(err))
            return ""

        return deserialize_null_terminated_string(data[hwcap_pointer:])



import unittest
from sotools.linker import resolve
from sotools.libraryset import Library

class LibraryTest(unittest.TestCase):
    def test_library_bad_object(self):
        with self.assertRaises(Exception):
            lib = Library.from_path("/tmp/lib.so")

    @unittest.skipIf(not resolve('libm.so.6') or not resolve('libc.so.6'), "No library to test with")
    def test_library_eq_library(self):
        sample = Library.from_path(resolve('libm.so.6'))
        other = Library.from_path(resolve('libm.so.6'))

        different = Library.from_path(resolve('libc.so.6'))

        self.assertEqual(sample, other)
        self.assertEqual(other, sample)
        self.assertNotEqual(sample, different)
        self.assertNotEqual(other, different)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_library_eq_any(self):
        sample = Library.from_path(resolve('libm.so.6'))

        self.assertNotEqual(sample, 'libm.so.6')
        self.assertNotEqual('libm.so.6', sample)



from copy import deepcopy
from pathlib import Path
from sotools import is_elf
from sotools.libraryset import Library, LibrarySet


class NotELFError(Exception):
    pass


def ldd(binary: str):
    """
    -> LibrarySet
    """

    path = Path(binary)

    if not is_elf(path):
        raise NotELFError

    executable = LibrarySet([Library.from_path(path)])
    libraries = deepcopy(executable)

    return LibrarySet(libraries.resolve() - executable)



from sotools.dl_cache.structure import BinaryStruct

# Appears as (uint32_t)-358342284 in glibc:/elf/cache.c
CACHE_EXTENSION_MAGIC = 3936625012


class CacheExtensionTag:
    TAG_GENERATOR = 0
    TAG_GLIBC_HWCAPS = 1
    COUNT = 2


class CacheExtensionSection(BinaryStruct):
    __structure__ = [
        ('tag', 'uint32_t'),
        ('flags', 'uint32_t'),
        ('offset', 'uint32_t'),
        ('size', 'uint32_t'),
    ]


class CacheExtension(BinaryStruct):
    __structure__ = [
        ('magic', 'uint32_t'),
        ('count', 'uint32_t'),
    ]


def cache_extension_sections(data: bytes):

    extension_header = CacheExtension.deserialize(data)

    header_size = BinaryStruct.sizeof(CacheExtension)
    section_size = BinaryStruct.sizeof(CacheExtensionSection)

    def parse_sections():
        for i in range(extension_header.count):
            offset = header_size + i * section_size
            yield CacheExtensionSection.deserialize(data[offset:])

    return list(parse_sections())



#!/bin/env python3

import sys
import logging
from argparse import ArgumentParser
from sotools.linker import resolve

DESCRIPTION = """This program will attempt to resolve an ELF file from a given shared object name. It allows to trace the attempts made by the linker to determine what shared object is resolved by what means."""
EPILOG = """Please report any mismatch between the dynamic linker and the output of this program to http://github.com/spoutn1k/python-sotools."""

PARSER = ArgumentParser(
    prog='sowhich',
    description=DESCRIPTION,
    epilog=EPILOG,
)

PARSER.add_argument(
    "soname",
    help="The library name to search for.",
)

PARSER.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="Trace resolving attempts while searching for the library",
)


def main():
    args = PARSER.parse_args()

    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
        )

    path = resolve(args.soname)
    if path:
        print(path)
        sys.exit(0)
    sys.exit(1)



"""
Library analysis and manipulation helpers
"""

from re import match
from os.path import realpath
from pathlib import Path
from typing import Set
from logging import debug

from sotools.libraryset import Library


def is_elf(path: Path) -> bool:
    """
    Returns a boolean if the file associated to the given path contains an ELF object.
    """
    try:
        with open(path, 'rb') as file:
            magic = file.read(4)

    except Exception:
        return False

    return magic == "\x7fELF".encode()


def library_links(shared_object: Library) -> Set[Path]:
    """
    This method resolves symbolic links that may exist and point to the
    library passed as an argument in the same directory as that library.

    Given the directory:
    lrwxrwxrwx. 1 root root   16 May 13  2019 libmpi.so -> libmpi.so.12.1.1
    lrwxrwxrwx. 1 root root   16 May 13  2019 libmpi.so.12 -> libmpi.so.12.1.1
    -rwxr-xr-x. 1 root root 2.7M May 13  2019 libmpi.so.12.1.1

    If any of those 3 files were to be passed as an argument, all would be
    returned.
    """
    if not isinstance(shared_object, Library):
        raise Exception(
            f"library_links: Wrong argument type: {type(shared_object)}")

    libname = Path(shared_object.binary_path).name

    # If no '.so' in the file name, return
    if not match(r'.*\.so.*', libname):
        debug("library_links: Error in format of %s", libname)
        return {Path(shared_object.binary_path)}

    cleared = set()
    prefix = libname.split('.so')[0]
    library_file = realpath(shared_object.binary_path)

    def _glob_links(prefix_):
        for file in list(Path(library_file).parent.glob(f"{prefix_}.so*")):
            if realpath(file) == library_file:
                cleared.add(file)

    _glob_links(prefix)

    # glib files are named as libc-2.33.so, but the links are named libc.so.x
    matches = match(r'(?P<prefix>lib[a-z_]+)-.+', prefix)
    if matches:
        _glob_links(matches.group('prefix'))

    # If we encounter a special case of symlink presenting as another library,
    # return the symlink and the shared object being pointed to.
    if shared_object.soname != libname:
        cleared.add(Path(shared_object.binary_path))
        cleared.add(Path(library_file))

    return cleared



import struct
import logging

DATATYPES = {
    # Datatypes used to describe the binary structure of the classes defined
    # to interpret the binary data. The first field is the byte size of the
    # member, the second field is the format string to pass to `struct.unpack`
    # when deserializing, and the last field is the default initializer to use.
    'uint8_t': (1, 'B', int),
    'int32_t': (4, 'i', int),
    'uint32_t': (4, 'I', int),
    'uint64_t': (8, 'Q', int),
}


class BinaryStruct:

    def sizeof(cls):
        # Check defined information
        hardcoded_size = getattr(cls, '_sizeof', None)
        structure = getattr(cls, '__structure__', None)

        # Return hardcoded size if it exists
        if hardcoded_size is not None:
            return hardcoded_size

        # Ensure the __structure__ attribute exists
        if structure is None:
            raise NotImplementedError(
                "Attempting to access size of class with no __structure__"
                f" field ({cls.__name__})")

        struct_size = 0

        # Resolve all members and cumulate their sizes
        for (attribute, type_) in structure:
            if attribute is None:
                if isinstance(type_, int):
                    struct_size += type_
                continue

            attr_size, _, _ = DATATYPES.get(type_)
            struct_size += attr_size

        return struct_size

    def __init__(self):
        for (attribute, type_) in self.__class__.__structure__:
            if attribute is None:
                continue

            _, _, default = DATATYPES.get(type_)

            setattr(self, attribute, default())

    @classmethod
    def deserialize(cls, data: bytes):
        if not isinstance(data, bytes):
            raise NotImplementedError(
                f"Unsupported value for deserialization buffer: {type(data)}")

        if getattr(cls, '__structure__', None) is None:
            raise NotImplementedError(
                "Attempting to deserialize class with no __structure__"
                f" field ({cls.__name__})")

        serialized = 0
        entry = cls()

        for (attribute, type_) in cls.__structure__:
            if attribute is None:
                serialized += type_
                continue

            size, format_, _ = DATATYPES.get(type_)

            try:
                value = struct.unpack(f'<{format_}',
                                      data[serialized:serialized + size])[0]
            except struct.error as err:
                raise Exception(
                    f"Error deserializing field {attribute} from object"
                    f" {cls.__name__}") from err

            setattr(entry, attribute, value)
            serialized = serialized + size

        return entry

    def __repr__(self):

        def _format_attributes():
            for (attribute, _) in getattr(self.__class__, '__structure__', []):
                if attribute is not None:
                    yield f"{attribute}={getattr(self, attribute, None)}"

        return f"{self.__class__.__name__}: " + ", ".join(_format_attributes())


def deserialize_null_terminated_string(data: bytes):
    terminator = data.find(0x0)

    if terminator == -1:
        logging.debug("Failed to find null byte in buffer")
        return ""

    return data[:terminator].decode(errors='replace')



#!/bin/env python3

import sys
import logging
from argparse import ArgumentParser
from sotools.ldd import ldd, NotELFError

DESCRIPTION = """List dynamic dependencies. This program will output a complete list of all the dynamic dependencies of the dynamic executable passed as an argument. This python version is safe to use on untrusted binaries."""
EPILOG = """Please report any mismatch between the dynamic linker and the output of this program to http://github.com/spoutn1k/python-sotools."""

PARSER = ArgumentParser(
    prog='ldd.py',
    description=DESCRIPTION,
    epilog=EPILOG,
)

PARSER.add_argument(
    "executable",
    help="Path to an executable to analyze.",
)

PARSER.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="Trace resolving attempts while searching for the dependencies",
)


def main():
    args = PARSER.parse_args()

    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
        )

    try:
        libs = ldd(args.executable)
    except NotELFError:
        print("\tnot a dynamic executable")
        sys.exit(1)

    print("\n".join(libs.ldd_format()))
    sys.exit(0)



"""
Implementation of the dynamic linker search algorithm
Rules in ld.so(8)
"""

import os
from typing import (
    List,
    Optional,
    Tuple,
)
from functools import lru_cache
from pathlib import Path
from sotools.dl_cache import cache_libraries
from sotools.dl_cache.flags import Flags
import logging

DEFAULT_PATHS = ['/lib', '/usr/lib', '/lib64', '/usr/lib64']


class LinkingError(Exception):
    pass


@lru_cache()
def _linker_path() -> Tuple[List[str], List[str]]:
    """
    Return linker search paths, in order
    Sourced from `man ld.so`
    """
    ld_library_path = filter(
        None,
        os.environ.get('LD_LIBRARY_PATH', "").split(':'),
    )

    return (ld_library_path, DEFAULT_PATHS)


def resolve(soname: str,
            rpath: Optional[List[str]] = None,
            runpath: Optional[List[str]] = None,
            arch_flags: Optional[Flags] = None) -> Optional[str]:
    """
    Get a path towards a library from a given soname.
    Implements system rules and takes the environment into account

    soname:     'key' to lookup for a library
    rpath:      rpath to use for this lookup
    runpath:    runpath to use for this lookup
    arch_flags: flags to look for; useful for 32bit libraries on 64bit systems
                See sotools.dl_cache.flags.Flags for info

    The method will return a resolved path for the given soname or None if
    no matching entry could be found.
    """

    found = Path()

    def _found() -> bool:
        """Check if a returned path corresponds to the soname"""
        return found != Path()

    def _valid(path: Path) -> bool:
        """Check a path is an existing directory"""
        return path.is_dir()

    def _search_paths(soname: str, paths: List[Path], reason: str) -> Path:
        """Search a list of paths and return the first match"""
        if paths:
            path_list_str = os.pathsep.join(map(lambda x: x.as_posix(), paths))
            logging.debug(f"search path={path_list_str}\t\t({reason or ''})")

        for dir_ in filter(_valid, paths):
            potential_lib = Path(dir_, soname)
            logging.debug(f"trying file={potential_lib.as_posix()}")
            if potential_lib.exists():
                return potential_lib

        return Path()

    rpath = list(map(Path, list(rpath or [])))
    runpath = list(map(Path, list(runpath or [])))
    cache_entries = cache_libraries(arch_flags=arch_flags)
    system_path = map(Path, _linker_path()[1])

    env_path, system_path = _linker_path()
    env_path = list(map(Path, env_path))
    system_path = list(map(Path, system_path))

    logging.debug(f"find library={soname}; searching")

    dynamic_paths = [
        (rpath, 'RPATH'),
        (env_path, 'LD_LIBRARY_PATH'),
        (runpath, 'RUNPATH'),
    ]

    # First, search the paths that are set by the user at run-time
    for paths, name in dynamic_paths:
        if not _found() and paths:
            found = _search_paths(soname, paths, name)

    # Query the cache for a match
    if not _found():
        logging.debug("search cache=/etc/ld.so.cache")
        if soname in cache_entries:
            found = Path(cache_entries[soname])

    default_paths = [(system_path, 'SYSTEM')]

    # Finally, search the hardcoded system paths
    for tuple_ in default_paths:
        if not _found():
            found = _search_paths(soname, *tuple_)

    if _found():
        logging.debug(f"found matching library={found}")
        return found.resolve()

    return None



import sys
import platform


class Flags:
    FLAG_ANY = -1
    FLAG_TYPE_MASK = 0x00ff
    FLAG_LIBC4 = 0x0000
    FLAG_ELF = 0x0001
    FLAG_ELF_LIBC5 = 0x0002
    FLAG_ELF_LIBC6 = 0x0003
    FLAG_REQUIRED_MASK = 0xff00
    FLAG_SPARC_LIB64 = 0x0100
    FLAG_IA64_LIB64 = 0x0200
    FLAG_X8664_LIB64 = 0x0300
    FLAG_S390_LIB64 = 0x0400
    FLAG_POWERPC_LIB64 = 0x0500
    FLAG_MIPS64_LIBN32 = 0x0600
    FLAG_MIPS64_LIBN64 = 0x0700
    FLAG_X8664_LIBX32 = 0x0800
    FLAG_ARM_LIBHF = 0x0900
    FLAG_AARCH64_LIB64 = 0x0a00
    FLAG_ARM_LIBSF = 0x0b00
    FLAG_MIPS_LIB32_NAN2008 = 0x0c00
    FLAG_MIPS64_LIBN32_NAN2008 = 0x0d00
    FLAG_MIPS64_LIBN64_NAN2008 = 0x0e00
    FLAG_RISCV_FLOAT_ABI_SOFT = 0x0f00
    FLAG_RISCV_FLOAT_ABI_DOUBLE = 0x1000

    _type_descriptions = {
        FLAG_LIBC4: "libc4",
        FLAG_ELF: "ELF",
        FLAG_ELF_LIBC5: "libc5",
        FLAG_ELF_LIBC6: "libc6",
    }

    _required_descriptions = {
        FLAG_SPARC_LIB64: "64bit",
        FLAG_IA64_LIB64: "IA-64",
        FLAG_X8664_LIB64: "x86-64",
        FLAG_S390_LIB64: "64bit",
        FLAG_POWERPC_LIB64: "64bit",
        FLAG_MIPS64_LIBN32: "N32",
        FLAG_MIPS64_LIBN64: "64bit",
        FLAG_X8664_LIBX32: "x32",
        FLAG_ARM_LIBHF: "hard-float",
        FLAG_AARCH64_LIB64: "AArch64",
        FLAG_ARM_LIBSF: "soft-float",
        FLAG_MIPS_LIB32_NAN2008: "nan2008",
        FLAG_MIPS64_LIBN32_NAN2008: "N32,nan2008",
        FLAG_MIPS64_LIBN64_NAN2008: "64bit,nan2008",
        FLAG_RISCV_FLOAT_ABI_SOFT: "soft-float",
        FLAG_RISCV_FLOAT_ABI_DOUBLE: "double-float",
    }

    @classmethod
    def description(cls, value: int):
        return ",".join([
            cls._type_descriptions.get(value & cls.FLAG_TYPE_MASK, "unknown"),
            cls._required_descriptions.get(value & cls.FLAG_REQUIRED_MASK,
                                           str(value & cls.FLAG_REQUIRED_MASK))
        ])

    @classmethod
    def is_64bits(cls, value: int):
        return value & cls.FLAG_REQUIRED_MASK in {
            cls.FLAG_SPARC_LIB64,
            cls.FLAG_IA64_LIB64,
            cls.FLAG_X8664_LIB64,
            cls.FLAG_S390_LIB64,
            cls.FLAG_POWERPC_LIB64,
            cls.FLAG_MIPS64_LIBN64,
            cls.FLAG_AARCH64_LIB64,
            cls.FLAG_MIPS64_LIBN64_NAN2008,
        }

    @classmethod
    def expected_flags(cls, executable: str = sys.executable):
        """
        Returns a integer value representing the expected flag value from the
        cache, or None if not found
        """
        # Return a machine name, from the 'uname' system call
        machine = platform.machine()
        # Use the provided path to get a bit number
        bits, _ = platform.architecture(executable)

        # Reference of expected flags from (machine x bits)
        # Found in glibc:/sysdeps/unix/sysv/linux/<ARCH>/dl-cache.h
        matches = dict([
            (('x86_64', '64bit'), cls.FLAG_X8664_LIB64 | cls.FLAG_ELF_LIBC6),
            (('x86_64', '32bit'), cls.FLAG_X8664_LIBX32 | cls.FLAG_ELF_LIBC6),
            (('ppc64le', '64bit'),
             cls.FLAG_POWERPC_LIB64 | cls.FLAG_ELF_LIBC6),
            (('arm', '32bit'), cls.FLAG_ARM_LIBHF | cls.FLAG_ELF_LIBC6),
            (('aarch64', '64bit'),
             cls.FLAG_AARCH64_LIB64 | cls.FLAG_ELF_LIBC6),
            (('aarch64_be', '64bit'),
             cls.FLAG_AARCH64_LIB64 | cls.FLAG_ELF_LIBC6),
        ])

        return matches.get((machine, bits))



"""
Defines a derived class of set() with shared objects, along with a few methods,
to facilitate handling large amounts of libraries
"""

import re
import logging
from pathlib import Path
from typing import Union

from sotools.util import flatten

from sotools.linker import resolve, LinkingError
from sotools.dl_cache import Flags

from elftools.common.exceptions import ELFError
from elftools.elf.elffile import ELFFile
from elftools.elf.dynamic import DynamicSection
from elftools.elf.gnuversions import (
    GNUVerDefSection,
    GNUVerNeedSection,
)


class Library:
    """
    Relevant ELF header fields used in the dynamic linking of libraries
    """

    @classmethod
    def from_path(cls, path: Union[str, Path]):
        library = cls()

        with open(path, 'rb') as file:
            try:
                for section in ELFFile(file).iter_sections():
                    if isinstance(section, GNUVerDefSection):
                        library.__parse_ver_def(section)
                    elif isinstance(section, GNUVerNeedSection):
                        library.__parse_ver_need(section)
                    elif isinstance(section, DynamicSection):
                        library.__parse_dynamic(section)
                library.binary_path = getattr(file, 'name', file)
            except (ELFError, AttributeError) as err:
                logging.error("Error parsing '%s' for ELF data: %s",
                              getattr(file, 'name', file), err)

        if not library.soname:
            library.soname = Path(path).name

        return library

    def __init__(self):
        self.soname = ''
        self.dyn_dependencies = set()
        self.required_versions = {}
        self.defined_versions = set()

        self.rpath = []
        self.runpath = []
        self.binary_path = None

    def __parse_dynamic(self, section):

        def __fetch_tags(id_):
            return list(
                filter(lambda x: x.entry.d_tag == id_, section.iter_tags()))

        tags = __fetch_tags('DT_SONAME')
        if len(tags) == 1:
            self.soname = tags[0].soname

        tags = __fetch_tags('DT_RPATH')
        if len(tags) == 1:
            self.rpath = tags[0].rpath.split(':')

        tags = __fetch_tags('DT_RUNPATH')
        if len(tags) == 1:
            self.runpath = tags[0].runpath.split(':')

        tags = __fetch_tags('DT_NEEDED')
        self.dyn_dependencies = {tag.needed for tag in tags}

    def __parse_ver_def(self, section):
        self.defined_versions = {
            next(v_iter).name
            for _, v_iter in section.iter_versions()
        }

    def __parse_ver_need(self, section):
        needed = {}

        for ver, v_iter in section.iter_versions():
            needed[ver.name] = {ver.name for ver in v_iter}

        self.required_versions = needed

    def __hash__(self):
        """
        hash method tying the ELFData object to the soname, to use in sets
        """
        return hash(self.soname)

    def __eq__(self, other):
        if isinstance(other, Library):
            return self.soname == other.soname and self.defined_versions == other.defined_versions
        return NotImplemented

    def __gt__(self, rhs):
        if isinstance(rhs, Library):
            return self.soname > rhs.soname
        return NotImplemented

    def __repr__(self):
        return f"'{self.soname}' from '{self.binary_path}'"


class LibrarySet(set):
    """
    Set-like object to collect Libray objects
    """

    @classmethod
    def create_from(cls, library_list):
        """
        -> LibrarySet[Library]
        Given a list of str or pathlib.Path, create a cf.libraries.LibrarySet
        with all the libraries and their dependencies resolved.
        """

        def _process(element):
            path = None

            if isinstance(element, Path):
                path = element.as_posix()
            elif isinstance(element, str):
                if '/' in element:
                    path = Path(element).as_posix()
                else:
                    path = resolve(element,
                                   rpath=cache.rpath,
                                   runpath=cache.runpath)
            else:
                raise Exception(
                    f"Wrong type for LibrarySet.create_from: {type(element)}")

            if not path:
                raise LinkingError(element)

            return path

        cache = LibrarySet()

        for path in map(_process, library_list):
            cache.add(Library.from_path(path))

        return cache.resolve()

    def add(self, elem):
        """
        -> None
        Wraps the default set.add method

        If a element with a given hash is added to a set, python will avoid
        copying it over if the set already contains an element with the same
        hash. This is an issue when adding HostLibraries/GuestLibraries
        depicting the same library, as the hash needs to be their soname.

        This method gets rid of matching libraries before calling set.add
        """
        if not isinstance(elem, Library):
            raise Exception(
                f"Adding object of incompatible type {type(elem)} to LibrarySet !"
            )

        conflict = list(filter(lambda x: hash(x) == hash(elem), self))

        if len(conflict) == 1:
            self.discard(conflict.pop())

        super().add(elem)

    @property
    def rpath(self):
        """
        -> list(str)
        Return a set of the libraries rpaths merged together
        """
        return flatten(map(lambda x: x.rpath, self))

    @property
    def runpath(self):
        """
        -> list(str)
        Return a set of the libraries runpaths merged together
        """
        return flatten(map(lambda x: x.runpath, self))

    @property
    def defined_versions(self):
        return set(flatten(map(lambda x: x.defined_versions, self)))

    @property
    def required_versions(self):
        return set(
            flatten(flatten(map(lambda x: x.required_versions.values(),
                                self))))

    @property
    def top_level(self):
        """
        -> LibrarySet, subset of self
        Returns a set of all libraries included in self that are not depended
        upon from another library in the set
        """
        return LibrarySet(self - self.required_libraries)

    @property
    def linkers(self):
        """
        -> LibrarySet, subset of self
        Returns a set of all linkers present in the set.
        The linker is identified by its static-ness and definition of private GLIB symbols.
        """
        return LibrarySet(
            filter(lambda x: not x.dyn_dependencies and x in self.glib, self))

    @property
    def glib(self):
        """
        -> LibrarySet, subset of self
        Returns a set with all libraries tied to the available libc,
        recognizable by the GLIBC_PRIVATE symbols. Using these with any
        other libc will trigger a symbol error.
        """

        def references_private(lib):
            return 'GLIBC_PRIVATE' in flatten(lib.required_versions.values(
            )) or 'GLIBC_PRIVATE' in lib.defined_versions

        return LibrarySet(filter(references_private, self))

    @property
    def required_libraries(self):
        """
        -> LibrarySet, subset of self
        Returns a set of all libraries included in self that are depended
        upon from another library in the set
        """
        sonames = set(flatten(map(lambda x: x.dyn_dependencies, self)))
        return LibrarySet(filter(lambda x: x.soname in sonames, self))

    @property
    def sonames(self):
        """
        -> set(str)
        Returns a set with the sonames of all the libraries in self
        """
        return set(map(lambda x: x.soname, self))

    @property
    def missing_libraries(self):
        """
        -> set(str)
        Returns a set with the sonames of all the dependencies of the set's
        libraries not present in self
        """
        req_sonames = set(flatten(map(lambda x: x.dyn_dependencies, self)))
        return req_sonames - self.sonames

    @property
    def outdated_libraries(self):
        """
        -> LibrarySet, subset of self
        Sonames of libraries that do not implement all the symbols expected by
        the other libraries
        """
        outdated = LibrarySet()

        for library in self:
            for _soname, required in library.required_versions.items():
                matches = set()

                for obj in self:
                    if obj.soname == _soname:
                        matches.add(obj)

                if len(matches) != 1:
                    continue

                dependency = matches.pop()

                if required > dependency.defined_versions:
                    outdated.add(dependency)

        return outdated

    @property
    def complete(self):
        """
        -> bool
        Returns True if all the dependencies are resolved
        """
        return (len(self.missing_libraries) == 0
                and self.required_versions.issubset(self.defined_versions))

    def find(self, soname):
        """
        -> Library or None
        Returns the matching library if found in self, else None
        """
        query = re.escape(soname)
        matches = set(filter(lambda x: re.match(query, x.soname), self))

        return matches.pop() if matches else None

    def resolve(self, rpath=None, runpath=None):
        """
        -> LibrarySet, superset of self
        will try to resolve all dynamic depedencies of the set's members, then
        add them to the returned set

        if the returned set complete() method returns False, a library cannot
        be found by e4s-cl
        """
        superset = LibrarySet(self)
        arch_flags = None

        # Assert all libraries are from the same architecture and store the
        # appropriate flag in arch_flag
        # Calculate all flags and boil them down in a set, filtering out Nones
        search_flags = {Flags.expected_flags(lib.binary_path) for lib in self}
        valid_flags = set(filter(None, search_flags))
        if len(valid_flags) == 1:
            arch_flags = list(valid_flags)[0]
        else:
            logging.debug(
                "Resolving dependencies of a set with mixed architectures (%s) !",
                ",".join(valid_flags))

        missing = superset.missing_libraries
        change = True

        while change:
            for soname in missing:
                path = resolve(soname,
                               rpath=superset.rpath,
                               runpath=superset.runpath,
                               arch_flags=arch_flags)
                logging.debug(f"Got path: {path}")

                if not path:
                    continue

                superset.add(Library.from_path(path))

            change = superset.missing_libraries != missing
            missing = superset.missing_libraries

        return superset

    def ldd_format(self):
        """
        -> list(str)
        """

        def line(soname: str):
            lib = self.find(soname)
            if lib and lib.binary_path:
                return "\t%(soname)s => %(binary_path)s" % lib.__dict__
            return f"\t{soname} => not found"

        lines = list(map(line, set.union(self.sonames,
                                         self.missing_libraries)))
        lines.sort()
        return lines



from pathlib import Path
import unittest
from sotools.libraryset import LibrarySet, Library
from sotools.linker import resolve


class LibrarySetTest(unittest.TestCase):

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_set(self):
        libset = LibrarySet([Library.from_path(resolve('libm.so.6'))])

        self.assertEqual(len(libset), 1)
        self.assertSetEqual(libset.sonames, {'libm.so.6'})

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_set_resolve(self):
        libset = LibrarySet([Library.from_path(resolve('libm.so.6'))])

        libset = libset.resolve()

        self.assertGreater(len(libset), 1)
        self.assertTrue(libset.sonames > {'libm.so.6', 'libc.so.6'})

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_set_missing(self):
        libset = LibrarySet([Library.from_path(resolve('libm.so.6'))])

        self.assertIn('libc.so.6', libset.missing_libraries)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_set_top(self):
        libset = LibrarySet([Library.from_path(resolve('libm.so.6'))])

        self.assertIn('libm.so.6', libset.top_level.sonames)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_create_from_soname(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        self.assertTrue(libset.complete)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_create_from_str_path(self):
        libset = LibrarySet.create_from([resolve('libm.so.6')])

        self.assertTrue(libset.complete)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_create_from_pathlib_path(self):
        libset = LibrarySet.create_from([Path(resolve('libm.so.6'))])

        self.assertTrue(libset.complete)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_get_glib(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        self.assertTrue(libset.glib)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_get_linkers(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        self.assertTrue(libset.linkers)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_get_outdated(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        self.assertFalse(libset.outdated_libraries)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_ldd(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        output = libset.ldd_format()

        for line in output:
            self.assertRegex(line, ".*.so.* => /.*")

    def test_ldd(self):
        lib = Library()
        lib.soname = 'libdummy.so.256'
        libset = LibrarySet([lib])

        output = libset.ldd_format()

        for line in output:
            self.assertRegex(line, ".*.so.* => not found")

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_escape_soname(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        self.assertFalse(libset.find('libc++'))

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_escape_soname(self):
        libset = LibrarySet.create_from(['libm.so.6'])

        self.assertFalse(libset.find('libc++'))

    def test_create_from_bad_input(self):
        with self.assertRaises(Exception):
            LibrarySet.create_from(range(10))

        with self.assertRaises(Exception):
            LibrarySet.create_from([None])

        with self.assertRaises(Exception):
            LibrarySet.create_from(['libnotalib.so'])

    def test_add_junk(self):
        with self.assertRaises(Exception):
            LibrarySet().add(1)

    @unittest.skipIf(not resolve('libm.so.6'), "No library to test with")
    def test_set(self):
        libset = LibrarySet()

        lib1 = Library()
        lib1.soname = "libnotalib.so"
        lib2 = Library()
        lib2.soname = "libnotalib.so"
        lib2.binary_path = "/tmp/notalib.so"

        libset.add(lib1)
        libset.add(lib2)

        self.assertEqual(len(libset), 1)
        saved_lib = libset.pop()
        self.assertEqual(saved_lib.binary_path, "/tmp/notalib.so")



import logging
from typing import List, Dict, Optional
from dataclasses import dataclass, field
from functools import lru_cache
from sotools.dl_cache.flags import Flags
from sotools.dl_cache.dl_cache import (_CacheHeader, _FileEntryNew)
from sotools.dl_cache.structure import (BinaryStruct,
                                        deserialize_null_terminated_string)
from sotools.dl_cache.extensions.hwcaps import (HWCAPSection,
                                                dl_cache_hwcap_extension)
from sotools.dl_cache.extensions.generator import GeneratorSection
from sotools.dl_cache.extensions import (cache_extension_sections,
                                         CacheExtensionTag)


def get_generator(data: bytes) -> Optional[str]:
    """
    Return the generator string from cache data, if the cache is recent enough
    to posess extensions
    """
    header = _CacheHeader.deserialize(data)
    data = data[header.offset:]

    if not header.extension_offset:
        logging.debug("Failed to retrieve generator: no extensions in cache")
        return None

    extensions = cache_extension_sections(data[header.extension_offset:])

    for section in extensions:
        if section.tag == CacheExtensionTag.TAG_GENERATOR:
            return GeneratorSection(section).string_value(data)
    return None


@dataclass(frozen=True)
class ResolvedEntry:
    key: str
    value: str
    flags: int
    hwcaps: str = ""


def _cache_libraries(data: bytes) -> List[ResolvedEntry]:
    """
    Return a list of ResolvedEntry objects with all references resolved
    """
    header = _CacheHeader.deserialize(data)
    data = data[header.offset:]
    extensions = []

    if header.extension_offset:
        extensions = cache_extension_sections(data[header.extension_offset:])

    def resolve_hwcap_values():
        for extension in extensions:
            if extension.tag == CacheExtensionTag.TAG_GLIBC_HWCAPS:
                yield HWCAPSection(extension).string_value(data)

    hwcap_string_values = list(resolve_hwcap_values())

    def _entries(data: bytes) -> list:
        """
        Generate a list of FileEntryXXX from the type defined in the header
        """
        header_size = BinaryStruct.sizeof(header.__class__)
        entry_type = header.__class__.entry_type
        entry_size = BinaryStruct.sizeof(entry_type)

        for index in range(header.nlibs):
            offset = header_size + index * entry_size
            entry = entry_type.deserialize(data[offset:offset + entry_size])
            yield entry

    def _lookup(entry):
        """
        Translate string references to strings
        """
        lookup = deserialize_null_terminated_string
        hwcap_entry_string = ""

        if isinstance(entry,
                      _FileEntryNew) and dl_cache_hwcap_extension(entry):
            index = entry.hwcap & ((1 << 32) - 1)
            if index < len(hwcap_string_values):
                hwcap_entry_string = hwcap_string_values[index]

        fields = dict(key=lookup(data[entry.key:]),
                      value=lookup(data[entry.value:]),
                      flags=entry.flags,
                      hwcaps=hwcap_entry_string)

        return ResolvedEntry(**fields)

    return list(map(_lookup, _entries(data)))


@dataclass(frozen=True)
class DynamicLinkerCache:
    file: str
    generator: Optional[str] = None
    entries: List[ResolvedEntry] = field(default_factory=list)


@lru_cache()
def _parse_cache(
        cache_file: str = "/etc/ld.so.cache") -> Optional[DynamicLinkerCache]:
    try:
        with open(cache_file, 'rb') as cache_file:
            cache_data = cache_file.read()
    except OSError as err:
        logging.error("Failed to open rtld cache: %s", str(err))
        return None

    try:
        entries = _cache_libraries(cache_data)
    except Exception as err:
        logging.error("rtdl cache parsing failed: %s", str(err))
        return None

    generator = get_generator(cache_data)

    fields = dict(file=cache_file, entries=entries, generator=generator)

    return DynamicLinkerCache(**fields)


def cache_libraries(cache_file: str = "/etc/ld.so.cache",
                    arch_flags: Optional[int] = None) -> Dict[str, str]:
    """
    Returns a dictionary with a curated list of the given cache file contents
    (/etc/ld.so.cache by default)

    cache: path towards a linker cache file
    flags: flag value to look for. A null value will return binaries matching
        the interpreter, a non-null value will be used to filter out mismatching
        entries. See Flags.expected_flags to create flag values

    Can be used to assume what libraries are installed on the system and where
    The keys are libraries' sonames; the values are the paths at which the
    corresponding shared object can be found

    The cache files may contain multiple entries for the same soname that differ
    in their flags, OS ABI and hardware capabilities. The dict returned
    contains one entry per soname, the most likely to be picked on the current
    system.

    See _cache_libraries for finer-grain control over the cache's contents
    """

    _arch_flags = arch_flags
    if _arch_flags is None:
        _arch_flags = Flags.expected_flags()

    cache = _parse_cache(cache_file)

    libs = cache.entries
    libs.reverse()

    def _generate_values():
        for entry in libs:
            if entry.flags == _arch_flags:
                yield (entry.key, entry.value)

    return dict(_generate_values())


def search_cache(soname: str,
                 cache_file: str = "/etc/ld.so.cache",
                 arch_flags: Optional[int] = None) -> Optional[str]:
    """
    Returns the best match for the given soname in the given cache matching
    the given flags

    soname: soname to match against the cache
    cache:  path towards a linker cache file
    flags:  flag value to look for. A null value will return binaries matching
        the interpreter, a non-null value will be used to filter out mismatching
        entries. See Flags.expected_flags to create flag values
    """

    _arch_flags = arch_flags
    if _arch_flags is None:
        _arch_flags = Flags.expected_flags()

    cache = _parse_cache(cache_file)

    for entry in cache.entries:
        # TODO hwcaps check, OS ABI check
        if entry.key == soname and entry.flags == _arch_flags:
            return entry.value

    return None



from sotools.dl_cache.structure import DATATYPES
from sotools.dl_cache.extensions import CacheExtensionSection


class GeneratorSection(CacheExtensionSection):

    def __init__(self, section):
        super().__init__()

        # Copy to self the fields defined by the class
        for attribute, type_ in CacheExtensionSection.__structure__:
            # Fetch the default for this attribute type
            _, _, default = DATATYPES.get(type_, int)

            # Access the field value then set it in self
            value = getattr(section, attribute, default())
            setattr(self, attribute, value)

    def string_value(self, data):
        return data[self.offset:self.offset + self.size].decode()



from sotools.dl_cache.structure import BinaryStruct


class _CacheType:
    """
    Used as an enum to differentiate cache types
    """
    UNKNOWN = 0x0
    NEW_FORMAT = 0x1
    OLD_FORMAT = 0x2


class _FileEntryOld(BinaryStruct):
    __structure__ = [('flags', 'int32_t'), ('key', 'uint32_t'),
                     ('value', 'uint32_t')]


class _FileEntryNew(BinaryStruct):
    __structure__ = [('flags', 'int32_t'), ('key', 'uint32_t'),
                     ('value', 'uint32_t'), ('osversion', 'uint32_t'),
                     ('hwcap', 'uint64_t')]


class _CacheHeaderOld(BinaryStruct):
    magic = "ld.so-1.7.0".encode()
    __structure__ = [(None, 12), ('nlibs', 'uint32_t')]
    entry_type = _FileEntryOld


class _CacheHeaderNew(BinaryStruct):
    __cachemagic_new = "glibc-ld.so.cache".encode()
    __cache_version = "1.1".encode()
    magic = __cachemagic_new + __cache_version
    __structure__ = [
        (None, 20),  # Cache magic and version string
        ('nlibs', 'uint32_t'),
        ('len_strings', 'uint32_t'),
        ('flags', 'uint8_t'),
        (None, 3),
        ('extension_offset', 'uint32_t'),
    ]
    _sizeof = 48
    entry_type = _FileEntryNew


def _format_error(*args, **kwargs):
    raise Exception("Data does not match a dynamic library cache")


def _cache_type(data: bytes):
    """ -> tuple[int]
    Determine the type of cache (from_CacheType) and the offset
    to start reading it from
    """

    if data[:len(_CacheHeaderNew.magic)] == _CacheHeaderNew.magic:
        return (_CacheType.NEW_FORMAT, 0)
    elif data[:len(_CacheHeaderOld.magic)] == _CacheHeaderOld.magic:
        # We do not have access to __alignof__, so search for magic
        offset = data.find(_CacheHeaderNew.magic)
        if offset != -1:
            return (_CacheType.NEW_FORMAT, offset)
        return (_CacheType.OLD_FORMAT, 0)

    return (_CacheType.UNKNOWN, 0)


class _CacheHeader:
    methods = {
        _CacheType.UNKNOWN: _format_error,
        _CacheType.NEW_FORMAT: _CacheHeaderNew.deserialize,
        _CacheType.OLD_FORMAT: _CacheHeaderOld.deserialize
    }

    @classmethod
    def deserialize(cls, data: bytes):
        cache_format, offset = _cache_type(data)

        header = cls.methods.get(cache_format, _format_error)(data[offset:])
        header.offset = offset

        return header
